###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
'''
mutable parameters for optimization:
    m.fs.system_recovery_target
'''
from pyomo.environ import ConcreteModel, Objective, Expression, Constraint, TransformationFactory, value, Param, Var
from pyomo.environ import units as pyunits
from pyomo.network import Arc

from idaes.core import FlowsheetBlock
from idaes.core.util.initialization import propagate_state
from idaes.core.util.scaling import calculate_scaling_factors

from watertap.examples.flowsheets.full_treatment_train.flowsheet_components import (desalination,
                                                                             feed_block,
                                                                             gypsum_saturation_index,
                                                                             translator_block,
                                                                             costing,
                                                                             financials, )
from watertap.examples.flowsheets.full_treatment_train.model_components import property_models
from watertap.examples.flowsheets.full_treatment_train.util import (solve_with_user_scaling,
                                                             check_dof)


desal_kwargs = {'has_desal_feed': False, 'is_twostage': False, 'has_ERD': True,
        'RO_type': '0D', 'RO_base': 'TDS', 'RO_level': 'detailed'}


def build_components(m, pretrt_type='NF', **kwargs):
    kwargs_desalination = {k: kwargs[k] for k in ('has_desal_feed', 'is_twostage', 'has_ERD',
                                                  'RO_type', 'RO_base', 'RO_level',)}

    desal_port = desalination.build_desalination(m, **kwargs_desalination)
    m.fs.s_tb_desal = Arc(source=m.fs.tb_pretrt_to_desal.outlet, destination=desal_port['in'])

    if pretrt_type == 'softening':
        property_models.build_prop(m, base='eNRTL')
    gypsum_saturation_index.build(m, section='desalination', pretrt_type=pretrt_type,  **kwargs)

    m.fs.RO.area.fix(80)
    m.fs.pump_RO.control_volume.properties_out[0].pressure.fix(60e5)

    if kwargs['is_twostage']:
        m.fs.RO2.area.fix(20)
        m.fs.pump_RO2.control_volume.properties_out[0].pressure.fix(90e5)

    # touch some properties used in optimization
    if kwargs['is_twostage']:
            product_water_sb = m.fs.mixer_permeate.mixed_state[0]
    else:
        if kwargs['RO_type'] == '0D':
            product_water_sb = m.fs.RO.permeate_side.properties_mixed[0]
        elif kwargs['RO_type'] == '1D':
            product_water_sb = m.fs.RO.mixed_permeate[0]

    feed_flow_vol = 0.0009769808  # value of feed flowrate using the seawater property package with 1 kg/s mass flowrate
    m.fs.system_recovery = Expression(
        expr=product_water_sb.flow_vol / feed_flow_vol)

    # need load factor from costing_param_block for annual_water_production
    financials.add_costing_param_block(m.fs)

    # RO recovery
    m.fs.RO_recovery = Var(initialize=0.5,
                           bounds=(0.01, 0.99),
                           doc='Total volumetric water recovery for RO')
    m.fs.eq_RO_recovery = Constraint(
        expr=m.fs.RO_recovery
             == product_water_sb.flow_vol / m.fs.tb_pretrt_to_desal.properties_out[0].flow_vol)

    # annual water production
    m.fs.annual_water_production = Expression(
        expr=pyunits.convert(product_water_sb.flow_vol, to_units=pyunits.m ** 3 / pyunits.year)
             * m.fs.costing_param.load_factor)
    costing.build_costing(m, module=financials, **kwargs)

    return desal_port


def build(m, **kwargs):
    """
    build an RO
    """
    assert not kwargs['has_desal_feed']
    property_models.build_prop(m, base='ion')
    feed_block.build_feed(m, base='ion')

    property_models.build_prop(m, base=kwargs['RO_base'])
    translator_block.build_tb(m, base_inlet='ion',
                              base_outlet=kwargs['RO_base'],
                              name_str='tb_pretrt_to_desal')
    m.fs.s_pretrt_tb = Arc(source=m.fs.feed.outlet, destination=m.fs.tb_pretrt_to_desal.inlet)

    property_models.build_prop(m, base='eNRTL')
    desal_port = build_components(m, **kwargs)


def scale(m, **kwargs):
    calculate_scaling_factors(m.fs.s_tb_desal)
    desalination.scale_desalination(m, **kwargs)


def initialize(m, **kwargs):
    propagate_state(m.fs.s_tb_desal)
    desalination.initialize_desalination(m, **kwargs)
    m.fs.desal_saturation.properties.initialize()


def report(m, **kwargs):
    m.fs.tb_pretrt_to_desal.report()
    desalination.display_desalination(m, **kwargs)
    print('desalination solubility index:', value(m.fs.desal_saturation.saturation_index))
    print('water recovery:', value(m.fs.system_recovery))
    costing.display_costing(m)


def set_optimization_components(m, system_recovery, **kwargs):
    '''
    adds max_saturation_index and system_recovery_target
    '''
    m.fs.pump_RO.control_volume.properties_out[0].pressure.unfix()
    m.fs.pump_RO.control_volume.properties_out[0].pressure.setlb(20e5)
    m.fs.pump_RO.control_volume.properties_out[0].pressure.setub(75e5)

    m.fs.RO.area.unfix()
    m.fs.RO.area.setlb(10)
    m.fs.RO.area.setub(300)

    if kwargs['RO_type'] == '0D':
        m.fs.RO.N_Re_io[0, 'in'].unfix()
    elif kwargs['RO_type'] == '1D':
        m.fs.RO.N_Re[0, 0].unfix()

    # Set lower bound for water flux at the RO outlet, based on a minimum net driving pressure, NDPmin
    m.fs.RO.NDPmin = Param(initialize=1e5, mutable=True, units=pyunits.Pa)
    if kwargs['RO_type'] == '0D':
        m.fs.RO.flux_mass_io_phase_comp[0, 'out', 'Liq', 'H2O'].setlb(value(m.fs.RO.A_comp[0, 'H2O']
                                                                            * m.fs.RO.dens_solvent
                                                                            * m.fs.RO.NDPmin))
    elif kwargs['RO_type'] == '1D':
        m.fs.RO.flux_mass_phase_comp[0, 1, 'Liq', 'H2O'].setlb(value(m.fs.RO.A_comp[0, 'H2O']
                                                                     * m.fs.RO.dens_solvent
                                                                     * m.fs.RO.NDPmin))

    # saturation index
    m.fs.max_saturation_index = Param(initialize=1.0, mutable=True)
    m.fs.eq_max_saturation_index_desal = Constraint(
        expr=m.fs.desal_saturation.saturation_index <= m.fs.max_saturation_index)

    m.fs.system_recovery_target = Param(initialize=system_recovery, mutable=True)
    m.fs.system_recovery_tol = Param(initialize=5e-3, mutable=True)
    m.fs.eq_system_recovery = Constraint(
        expr=(m.fs.system_recovery_target,
                m.fs.system_recovery,
                m.fs.system_recovery_target+m.fs.system_recovery_tol))

    # set objective
    m.fs.objective = Objective(expr=m.fs.costing.LCOW)


def set_up_optimization(m, system_recovery=0.50, **kwargs):
    set_optimization_components(m, system_recovery, **kwargs)
    check_dof(m, 3)


def optimize(m):
    solve_with_user_scaling(m, tee=False, fail_flag=True)


def solve_flowsheet(**desal_kwargs):
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    build(m, **desal_kwargs)
    TransformationFactory("network.expand_arcs").apply_to(m)

    # scale
    calculate_scaling_factors(m.fs.tb_pretrt_to_desal)
    scale(m, **desal_kwargs)
    calculate_scaling_factors(m)

    # initialize
    m.fs.feed.initialize()
    propagate_state(m.fs.s_pretrt_tb)
    optarg = {'nlp_scaling_method': 'user-scaling'}
    m.fs.tb_pretrt_to_desal.initialize(optarg=optarg)
    initialize(m, **desal_kwargs)

    check_dof(m)
    solve_with_user_scaling(m, tee=False, fail_flag=True)

    # report
    report(m, **desal_kwargs)

    return m


def optimize_flowsheet(system_recovery=0.50, **kwargs):
    m = solve_flowsheet(**kwargs)
    set_up_optimization(m, system_recovery=system_recovery, **kwargs)
    optimize(m)

    report(m, **kwargs)

    return m


if __name__ == "__main__":
    import sys
    if len(sys.argv) == 1:
        m = solve_flowsheet(**desal_kwargs)
    else:
        m = optimize_flowsheet(system_recovery=float(sys.argv[1]), **desal_kwargs)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

import pytest

from pyomo.environ import ConcreteModel, Var, Constraint, Block, SolverFactory

from idaes.core.util import get_solver
from watertap.core.util.initialization import (check_dof,
                                               assert_degrees_of_freedom,
                                               assert_no_degrees_of_freedom,
                                               check_solve,
                                               generate_initialization_perturbation,
                                               print_initialization_perturbation,
                                               assert_no_initialization_perturbation)
import idaes.logger as idaeslog

__author__ = "Adam Atia"

_log = idaeslog.getLogger(__name__)

# Set up solver
solver = get_solver()

class TestCheckDOF:
    @pytest.fixture(scope="class")
    def m(self):
        m = ConcreteModel()
        m.a = Var()
        m.b = Var()
        m.abcon = Constraint(rule=m.a + m.b == 10)
        return m

    @pytest.mark.unit
    def test_expected(self, m):
        check_dof(m, fail_flag=False, expected_dof=1)
        check_dof(m, fail_flag=True, expected_dof=1)
        assert_degrees_of_freedom(m, 1)

    @pytest.mark.unit
    def test_more_expected(self, m):
        check_dof(m, fail_flag=False, expected_dof=3)
        msg = r"Unexpected degrees of freedom: Degrees of freedom on unknown = 1. Expected 3. Unfix 2 variable\(s\)"
        with pytest.raises(ValueError, match=msg):
            check_dof(m, fail_flag=True, expected_dof=3)
        with pytest.raises(ValueError, match=msg):
            assert_degrees_of_freedom(m, 3)

    @pytest.mark.unit
    def test_less_expected(self, m):
        check_dof(m, fail_flag=False, expected_dof=-1)
        msg = r"Unexpected degrees of freedom: Degrees of freedom on unknown = 1. Expected -1. Fix 2 variable\(s\)"
        with pytest.raises(ValueError, match=msg):
            check_dof(m, fail_flag=True, expected_dof=-1)
        with pytest.raises(ValueError, match=msg):
            assert_degrees_of_freedom(m, -1)

    @pytest.mark.unit
    def test_zero_expected(self, m):
        # check_dof should pass since fail_flag=False produces warning for DOF!=0
        check_dof(m, fail_flag=False)
        msg = r"Non-zero degrees of freedom: Degrees of freedom on unknown = 1. Fix 1 more variable\(s\)"
        # Verify error is raised since DOF!=0
        with pytest.raises(ValueError, match=msg):
            check_dof(m, fail_flag=True)
        with pytest.raises(ValueError, match=msg):
            assert_no_degrees_of_freedom(m)

    @pytest.mark.unit
    def test_zero(self, m):
        m.a.fix(5)
        # check should pass since DOF=0
        check_dof(m, fail_flag=True)
        check_dof(m, fail_flag=True, expected_dof=0)
        assert_no_degrees_of_freedom(m)

        m.a.unfix()


class TestCheckSolve:
    @pytest.fixture(scope="class")
    def m(self):
        m = ConcreteModel()
        m.a = Var()
        m.acon = Constraint(rule=m.a >= 10)
        m.bcon = Constraint(rule=m.a == 5)
        return m

    @pytest.mark.unit
    def test_failure(self, m):
        results = solver.solve(m)
        # check_solve should pass since fail_flag=False and only warning will be produced
        check_solve(results, logger=_log, fail_flag=False)
        # expect the solve to fail and raise error
        with pytest.raises(ValueError, match="The solver failed to converge to an optimal solution. This suggests that the "
                                             "user provided infeasible inputs or that the model is poorly scaled."):
            check_solve(results, logger=_log, fail_flag=True)

    @pytest.mark.unit
    def test_failure_checkpoint(self, m):
        results = solver.solve(m)
        # check_solve should pass since fail_flag=False and only warning will be produced
        check_solve(results, checkpoint='test', logger=_log, fail_flag=False)
        # expect the solve to fail and raise error
        with pytest.raises(ValueError, match="test failed. The solver failed to converge to an optimal solution. "
                "This suggests that the user provided infeasible inputs or that the model is poorly scaled."):
            check_solve(results, checkpoint='test', logger=_log, fail_flag=True)
    
    @pytest.mark.unit
    def test_success(self, m):
        m.acon.deactivate()

        results = solver.solve(m)
        # both check_solve's should pass
        check_solve(results, logger=_log, fail_flag=False)
        check_solve(results, logger=_log, fail_flag=True)

        m.acon.activate()


class TestPerturbationHelper:
    @pytest.fixture(scope="class")
    def b(self):
        b = Block(concrete=True)
        b.x = Var(bounds=(1e-8, None), initialize=1e-7)
        b.y = Var(bounds=(1e+1, 1e+2), initialize=1e+3)

        b.z = Var(bounds=(0., 1e-8), initialize=1e-20)
        b.z.fix()

        b.w = Var(bounds=(None, 1), initialize=0.5)

        return b

    @pytest.mark.unit
    def test_generate_initialization_perturbation(self, b):
        r = list(generate_initialization_perturbation(b))
        assert r[0][0].name == 'x'
        assert r[1][0].name == 'y'

        assert r[0][1] == 1e-7
        assert r[1][1] == 1e+3

        assert r[0][2] == 1e-2
        assert r[1][2] == 99.100000989

        r = list(generate_initialization_perturbation(b, bound_relax_factor=0.))
        assert r[0][0].name == 'x'
        assert r[1][0].name == 'y'

        assert r[0][2] == 1.000001e-2
        assert r[1][2] == 99.1

        r = list(generate_initialization_perturbation(b, bound_relax_factor=0., bound_frac=1e-3))
        assert r[0][0].name == 'x'
        assert r[1][0].name == 'y'

        assert r[0][2] == 1.000001e-2
        assert r[1][2] == 99.91

        r = list(generate_initialization_perturbation(b, bound_relax_factor=0., bound_frac=1e-3, bound_push=1e-3))
        assert r[0][0].name == 'x'
        assert r[1][0].name == 'y'

        assert r[0][2] == 1.00001e-3
        assert r[1][2] == 99.91

        r = list(generate_initialization_perturbation(b, bound_relax_factor=0., bound_frac=1e-3, bound_push=1e-10))
        assert r[0][0].name == 'y'
        assert r[0][2] == 100.-1e-8

        r = list(generate_initialization_perturbation(b, bound_push=1e-6))
        assert r[0][0].name == 'x'
        assert r[1][0].name == 'y'

        assert r[0][2] == 1.e-6
        assert r[1][2] == 99.999900999999

    @pytest.mark.unit
    def test_print_initialization_perturbation(self, b, capsys):
        print_initialization_perturbation(b, 1e-2, 1e-2, 1e-8, True)
        captured = capsys.readouterr()
        assert captured.out == \
"""IPOPT will move scaled initial value for variable x from 1.000000e-07 to 1.000000e-02
IPOPT will move scaled initial value for variable y from 1.000000e+03 to 9.910000e+01
"""

    @pytest.mark.unit
    def test_assert_no_initialization_perturbation1(self, b):
        with pytest.raises(ValueError, match=\
                "IPOPT will move scaled initial value for variable x from 1.000000e-07 to 1.000000e-02"):
            assert_no_initialization_perturbation(b)

    @pytest.mark.unit
    def test_assert_no_initialization_perturbation2(self, b):
        optarg = {'bound_push' : 1e-10}
        b.y.value = 5e+1
        assert_no_initialization_perturbation(b, optarg=optarg)

    @pytest.mark.unit
    def test_assert_no_initialization_perturbation3(self, b):
        solver = get_solver()
        solver.options['bound_push'] = 1e-10
        b.y.value = 5e+1
        assert_no_initialization_perturbation(b, solver=solver)

    @pytest.mark.unit
    def test_assert_no_initialization_perturbation4(self, b):
        with pytest.raises(ValueError, match="Supply a solver or optarg, not both"):
            assert_no_initialization_perturbation(b, solver=b, optarg=b)

    @pytest.mark.unit
    def test_assert_no_initialization_perturbation5(self, b):
        with pytest.raises(ValueError, match="Solver cbc is not supported"):
            assert_no_initialization_perturbation(b, solver=SolverFactory('cbc'))



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

import pytest
from watertap.property_models.tests.property_test_harness import \
    (PropertyTestHarness, PropertyRegressionTest)
from pyomo.environ import ConcreteModel, value
from idaes.core import FlowsheetBlock
import idaes.core.util.scaling as iscale
from pyomo.util.check_units import assert_units_consistent
from watertap.examples.flowsheets.full_treatment_train.util import solve_with_user_scaling, check_dof
import watertap.examples.flowsheets.full_treatment_train.model_components.seawater_ion_prop_pack as property_seawater_ions
import watertap.examples.flowsheets.full_treatment_train.model_components.seawater_salt_prop_pack as property_seawater_salts

# -----------------------------------------------------------------------------
@pytest.mark.component
def test_property_seawater_ions():
    m = ConcreteModel()

    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = property_seawater_ions.PropParameterBlock()
    m.fs.stream = m.fs.properties.build_state_block([0], default={})

    # specify
    feed_flow_mass = 1
    feed_mass_frac = {'Na': 11122e-6,
                      'Ca': 382e-6,
                      'Mg': 1394e-6,
                      'SO4': 2136e-6,
                      'Cl': 20300e-6}
    m.fs.stream[0].flow_mass_phase_comp['Liq', 'H2O'].fix(
        feed_flow_mass * (1 - sum(x for x in feed_mass_frac.values())))
    for s in feed_mass_frac:
        m.fs.stream[0].flow_mass_phase_comp['Liq', s].fix(feed_flow_mass * feed_mass_frac[s])
    m.fs.stream[0].temperature.fix(273.15 + 25)
    m.fs.stream[0].pressure.fix(101325)

    m.fs.stream[0].mass_frac_phase_comp
    m.fs.stream[0].flow_mol_phase_comp

    # scaling
    m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1, index=('Liq', 'H2O'))
    m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e2, index=('Liq', 'Na'))
    m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e4, index=('Liq', 'Ca'))
    m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e3, index=('Liq', 'Mg'))
    m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e3, index=('Liq', 'SO4'))
    m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e2, index=('Liq', 'Cl'))
    iscale.calculate_scaling_factors(m.fs)

    # checking state block
    assert_units_consistent(m)
    check_dof(m)

    # initialize
    m.fs.stream.initialize(optarg={'nlp_scaling_method': 'user-scaling'})

    # solve
    solve_with_user_scaling(m)

    # check values
    assert value(m.fs.stream[0].mass_frac_phase_comp['Liq', 'H2O']) == pytest.approx(0.9647, rel=1e-3)
    assert value(m.fs.stream[0].flow_mol_phase_comp['Liq', 'Na']) == pytest.approx(0.4838, rel=1e-3)
    assert value(m.fs.stream[0].flow_mol_phase_comp['Liq', 'Ca']) == pytest.approx(9.531e-3, rel=1e-3)

class TestPropertySeawaterIons(PropertyTestHarness):
    def configure(self):
        self.prop_pack = property_seawater_ions.PropParameterBlock
        self.param_args = {}
        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1,
                             ('flow_mass_phase_comp', ('Liq', 'Na')): 1e2,
                             ('flow_mass_phase_comp', ('Liq', 'Ca')): 1e4,
                             ('flow_mass_phase_comp', ('Liq', 'Mg')): 1e3,
                             ('flow_mass_phase_comp', ('Liq', 'SO4')): 1e3,
                             ('flow_mass_phase_comp', ('Liq', 'Cl')): 1e2}
        self.stateblock_statistics = {'number_variables': 27,
                                      'number_total_constraints': 19,
                                      'number_unused_variables': 2,  # pressure and temperature are unused
                                      'default_degrees_of_freedom': 6}  # 8 state vars, pressure and temp are not active
        self.default_solution = {('mass_frac_phase_comp', ('Liq', 'H2O')): 0.9647,
                                 ('flow_vol', None): 1e-3,
                                 ('flow_mol_phase_comp', ('Liq', 'Na')): 0.4838,
                                 ('flow_mol_phase_comp', ('Liq', 'Ca')): 9.531e-3,
                                 ('enth_flow', None): 1.05e5}


class TestPropertySeawaterSalts(PropertyTestHarness):
    def configure(self):
        self.prop_pack = property_seawater_salts.PropParameterBlock
        self.param_args = {}
        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1,
                             ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1e2,
                             ('flow_mass_phase_comp', ('Liq', 'CaSO4')): 1e3,
                             ('flow_mass_phase_comp', ('Liq', 'MgSO4')): 1e3,
                             ('flow_mass_phase_comp', ('Liq', 'MgCl2')): 1e3}
        self.stateblock_statistics = {'number_variables': 18,
                                      'number_total_constraints': 11,
                                      'number_unused_variables': 2,  # pressure and temperature are unused
                                      'default_degrees_of_freedom': 5}  # 7 state vars, pressure and temp are not active
        self.default_solution = {('mass_frac_phase_comp', ('Liq', 'H2O')): 0.9647,
                                 ('flow_vol', None): 1e-3,
                                 ('flow_mol_phase_comp', ('Liq', 'NaCl')): 0.4838,
                                 ('flow_mol_phase_comp', ('Liq', 'CaSO4')): 9.531e-3,
                                 ('flow_mol_phase_comp', ('Liq', 'MgSO4')): 1.270e-2,
                                 ('flow_mol_phase_comp', ('Liq', 'MgCl2')): 4.465e-2,
                                 ('enth_flow', None): 1.05e5}



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

"""
    This test is to establish that the core chemistry packages in IDAES solve
    a simple water dissociation problem and return the correct pH value.
"""
import enum

# Importing testing libraries
import pytest

# Importing the object for units from pyomo
from pyomo.environ import units as pyunits

# Imports from idaes core
from idaes.core import AqueousPhase
from idaes.core.components import Solvent, Solute, Cation, Anion
from idaes.core.phases import PhaseType as PT

# Imports from idaes generic models
import idaes.generic_models.properties.core.pure.Perrys as Perrys
from idaes.generic_models.properties.core.state_definitions import FTPx
from idaes.generic_models.properties.core.eos.ideal import Ideal

# Importing the enum for concentration unit basis used in the 'get_concentration_term' function
from idaes.generic_models.properties.core.generic.generic_reaction import (
    ConcentrationForm,
)

# Import the object/function for heat of reaction
from idaes.generic_models.properties.core.reactions.dh_rxn import constant_dh_rxn

# Import safe log power law equation
from idaes.generic_models.properties.core.reactions.equilibrium_forms import (
    log_power_law_equil,
)

# Import k-value functions
from idaes.generic_models.properties.core.reactions.equilibrium_constant import (
    gibbs_energy,
    van_t_hoff,
)

# Import specific pyomo objects
from pyomo.environ import (
    ConcreteModel,
    SolverStatus,
    TerminationCondition,
    value,
    Suffix,
)

from idaes.core.util import scaling as iscale

# Import pyomo methods to check the system units
from pyomo.util.check_units import assert_units_consistent

# Import idaes methods to check the model during construction
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    fixed_variables_set,
    activated_constraints_set,
    number_variables,
    number_total_constraints,
    number_unused_variables,
)

# Import the idaes objects for Generic Properties and Reactions
from idaes.generic_models.properties.core.generic.generic_property import (
    GenericParameterBlock,
)
from idaes.generic_models.properties.core.generic.generic_reaction import (
    GenericReactionParameterBlock,
)

# Import the idaes object for the EquilibriumReactor unit model
from idaes.generic_models.unit_models.equilibrium_reactor import EquilibriumReactor

# Import the core idaes objects for Flowsheets and types of balances
from idaes.core import FlowsheetBlock

# Import log10 function from pyomo
from pyomo.environ import log10


__authors__ = [
    "Austin Ladshaw",
    "Ludovico Bianchi",
    "Dan Gunter",
]
__author__ = __authors__[0]


class Variant(str, enum.Enum):
    equilibrium = "equilibrium"
    inherent = "inherent"

    def __str__(self):
        return f'{self.__class__.__name__}.{self.name}'

    @property
    def is_equilibrium(self):
        return self is Variant.equilibrium


@pytest.fixture(
    scope="module",
    params=[
        Variant.equilibrium,
        Variant.inherent
    ],
    ids=str
)
def variant(request) -> Variant:
    return request.param


@pytest.fixture(scope="module")
def base_units():
    return {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    }


@pytest.fixture(scope="module")
def solver():
    s = get_solver()
    s.options["max_iter"] = 200
    return s


@pytest.fixture(scope="module")
def thermo_config(base_units):
    # Configuration dictionary
    data = {
        "components": {
            "H2O": {
                "type": Solvent,
                # Define the methods used to calculate the following properties
                "dens_mol_liq_comp": Perrys,
                "enth_mol_liq_comp": Perrys,
                "cp_mol_liq_comp": Perrys,
                "entr_mol_liq_comp": Perrys,
                # Parameter data is always associated with the methods defined above
                "parameter_data": {
                    "mw": (18.0153, pyunits.g / pyunits.mol),
                    "pressure_crit": (220.64e5, pyunits.Pa),
                    "temperature_crit": (647, pyunits.K),
                    # Comes from Perry's Handbook:  p. 2-98
                    "dens_mol_liq_comp_coeff": {
                        "1": (5.459, pyunits.kmol * pyunits.m ** -3),
                        "2": (0.30542, pyunits.dimensionless),
                        "3": (647.13, pyunits.K),
                        "4": (0.081, pyunits.dimensionless),
                    },
                    "enth_mol_form_liq_comp_ref": (-285.830, pyunits.kJ / pyunits.mol),
                    "enth_mol_form_vap_comp_ref": (0, pyunits.kJ / pyunits.mol),
                    # Comes from Perry's Handbook:  p. 2-174
                    "cp_mol_liq_comp_coeff": {
                        "1": (2.7637e5, pyunits.J / pyunits.kmol / pyunits.K),
                        "2": (-2.0901e3, pyunits.J / pyunits.kmol / pyunits.K ** 2),
                        "3": (8.125, pyunits.J / pyunits.kmol / pyunits.K ** 3),
                        "4": (-1.4116e-2, pyunits.J / pyunits.kmol / pyunits.K ** 4),
                        "5": (9.3701e-6, pyunits.J / pyunits.kmol / pyunits.K ** 5),
                    },
                    "cp_mol_ig_comp_coeff": {
                        "A": (30.09200, pyunits.J / pyunits.mol / pyunits.K),
                        "B": (
                            6.832514,
                            pyunits.J
                            * pyunits.mol ** -1
                            * pyunits.K ** -1
                            * pyunits.kiloK ** -1,
                        ),
                        "C": (
                            6.793435,
                            pyunits.J
                            * pyunits.mol ** -1
                            * pyunits.K ** -1
                            * pyunits.kiloK ** -2,
                        ),
                        "D": (
                            -2.534480,
                            pyunits.J
                            * pyunits.mol ** -1
                            * pyunits.K ** -1
                            * pyunits.kiloK ** -3,
                        ),
                        "E": (
                            0.082139,
                            pyunits.J
                            * pyunits.mol ** -1
                            * pyunits.K ** -1
                            * pyunits.kiloK ** 2,
                        ),
                        "F": (-250.8810, pyunits.kJ / pyunits.mol),
                        "G": (223.3967, pyunits.J / pyunits.mol / pyunits.K),
                        "H": (0, pyunits.kJ / pyunits.mol),
                    },
                    "entr_mol_form_liq_comp_ref": (
                        69.95,
                        pyunits.J / pyunits.K / pyunits.mol,
                    ),
                    "pressure_sat_comp_coeff": {
                        "A": (4.6543, None),  # [1], temperature range 255.9 K - 373 K
                        "B": (1435.264, pyunits.K),
                        "C": (-64.848, pyunits.K),
                    },
                },
                # End parameter_data
            },
            "H_+": {
                "type": Cation,
                "charge": 1,
                # Define the methods used to calculate the following properties
                "dens_mol_liq_comp": Perrys,
                "enth_mol_liq_comp": Perrys,
                "cp_mol_liq_comp": Perrys,
                "entr_mol_liq_comp": Perrys,
                # Parameter data is always associated with the methods defined above
                "parameter_data": {
                    "mw": (1.00784, pyunits.g / pyunits.mol),
                    "dens_mol_liq_comp_coeff": {
                        "1": (5.459, pyunits.kmol * pyunits.m ** -3),
                        "2": (0.30542, pyunits.dimensionless),
                        "3": (647.13, pyunits.K),
                        "4": (0.081, pyunits.dimensionless),
                    },
                    "enth_mol_form_liq_comp_ref": (-230.000, pyunits.kJ / pyunits.mol),
                    "cp_mol_liq_comp_coeff": {
                        "1": (2.7637e5, pyunits.J / pyunits.kmol / pyunits.K),
                        "2": (-2.0901e3, pyunits.J / pyunits.kmol / pyunits.K ** 2),
                        "3": (8.125, pyunits.J / pyunits.kmol / pyunits.K ** 3),
                        "4": (-1.4116e-2, pyunits.J / pyunits.kmol / pyunits.K ** 4),
                        "5": (9.3701e-6, pyunits.J / pyunits.kmol / pyunits.K ** 5),
                    },
                    "entr_mol_form_liq_comp_ref": (
                        -10.75,
                        pyunits.J / pyunits.K / pyunits.mol,
                    ),
                },
                # End parameter_data
            },
            "OH_-": {
                "type": Anion,
                "charge": -1,
                # Define the methods used to calculate the following properties
                "dens_mol_liq_comp": Perrys,
                "enth_mol_liq_comp": Perrys,
                "cp_mol_liq_comp": Perrys,
                "entr_mol_liq_comp": Perrys,
                # Parameter data is always associated with the methods defined above
                "parameter_data": {
                    "mw": (17.008, pyunits.g / pyunits.mol),
                    "dens_mol_liq_comp_coeff": {
                        "1": (5.459, pyunits.kmol * pyunits.m ** -3),
                        "2": (0.30542, pyunits.dimensionless),
                        "3": (647.13, pyunits.K),
                        "4": (0.081, pyunits.dimensionless),
                    },
                    "enth_mol_form_liq_comp_ref": (-230.000, pyunits.kJ / pyunits.mol),
                    "cp_mol_liq_comp_coeff": {
                        "1": (2.7637e5, pyunits.J / pyunits.kmol / pyunits.K),
                        "2": (-2.0901e3, pyunits.J / pyunits.kmol / pyunits.K ** 2),
                        "3": (8.125, pyunits.J / pyunits.kmol / pyunits.K ** 3),
                        "4": (-1.4116e-2, pyunits.J / pyunits.kmol / pyunits.K ** 4),
                        "5": (9.3701e-6, pyunits.J / pyunits.kmol / pyunits.K ** 5),
                    },
                    "entr_mol_form_liq_comp_ref": (
                        -10.75,
                        pyunits.J / pyunits.K / pyunits.mol,
                    ),
                },
                # End parameter_data
            },
        },
        # End Component list
        "phases": {
            "Liq": {"type": AqueousPhase, "equation_of_state": Ideal},
        },
        "state_definition": FTPx,
        "state_bounds": {
            "flow_mol": (0, 50, 100),
            "temperature": (273.15, 300, 650),
            "pressure": (5e4, 1e5, 1e6),
        },
        "pressure_ref": 1e5,
        "temperature_ref": 300,
        "base_units": base_units,
        "inherent_reactions": {
            "H2O_Kw": {
                "stoichiometry": {
                    ("Liq", "H2O"): -1,
                    ("Liq", "H_+"): 1,
                    ("Liq", "OH_-"): 1,
                },
                "heat_of_reaction": constant_dh_rxn,
                "equilibrium_constant": van_t_hoff,
                "equilibrium_form": log_power_law_equil,
                "concentration_form": ConcentrationForm.moleFraction,
                "parameter_data": {
                    "dh_rxn_ref": (55.830, pyunits.J / pyunits.mol),
                    "k_eq_ref": (10 ** -14 / 55.2 / 55.2, pyunits.dimensionless),
                    "T_eq_ref": (298, pyunits.K),
                    # By default, reaction orders follow stoichiometry
                    #    manually set reaction order here to override
                    "reaction_order": {
                        ("Liq", "H2O"): 0,
                        ("Liq", "H_+"): 1,
                        ("Liq", "OH_-"): 1,
                    },
                }
                # End parameter_data
            }
            # End R1
        }
    }
    return data


# Define the reaction_config for water dissociation
@pytest.fixture(scope="module")
def water_reaction_config(base_units):
    return {
        "base_units": base_units,
        "equilibrium_reactions": {
            "H2O_Kw": {
                "stoichiometry": {
                    ("Liq", "H2O"): -1,
                    ("Liq", "H_+"): 1,
                    ("Liq", "OH_-"): 1,
                },
                "heat_of_reaction": constant_dh_rxn,
                "equilibrium_constant": van_t_hoff,
                "equilibrium_form": log_power_law_equil,
                "concentration_form": ConcentrationForm.moleFraction,
                "parameter_data": {
                    "dh_rxn_ref": (55.830, pyunits.J / pyunits.mol),
                    "k_eq_ref": (10 ** -14 / 55.2 / 55.2, pyunits.dimensionless),
                    "T_eq_ref": (298, pyunits.K),
                    # By default, reaction orders follow stoichiometry
                    #    manually set reaction order here to override
                    "reaction_order": {
                        ("Liq", "H2O"): 0,
                        ("Liq", "H_+"): 1,
                        ("Liq", "OH_-"): 1,
                    },
                }
                # End parameter_data
            }
            # End R1
        }
        # End equilibrium_reactions
    }
    # End reaction_config definition

def _get_without_inherent_reactions(config_dict: dict, key="inherent_reactions") -> dict:
    data = dict(config_dict)
    del data[key]
    return data

class TestPureWater:


    @pytest.fixture
    def model(self, thermo_config, variant: Variant, water_reaction_config):
        if variant.is_equilibrium:
            thermo_config = _get_without_inherent_reactions(thermo_config)
        model = ConcreteModel()
        model.fs = FlowsheetBlock(default={"dynamic": False})
        model.fs.thermo_params = GenericParameterBlock(default=thermo_config)
        print(water_reaction_config)
        model.fs.rxn_params = GenericReactionParameterBlock(
            default={"property_package": model.fs.thermo_params, **water_reaction_config}
        )
        model.fs.unit = EquilibriumReactor(
            default={
                "property_package": model.fs.thermo_params,
                "reaction_package": model.fs.rxn_params,
                "has_rate_reactions": False,
                "has_equilibrium_reactions": variant.is_equilibrium,
                "has_heat_transfer": False,
                "has_heat_of_reaction": False,
                "has_pressure_change": False,
            }
        )

        model.fs.unit.inlet.mole_frac_comp[0, "H_+"].fix(0.0)
        model.fs.unit.inlet.mole_frac_comp[0, "OH_-"].fix(0.0)
        model.fs.unit.inlet.mole_frac_comp[0, "H2O"].fix(1.0)
        model.fs.unit.inlet.pressure.fix(101325.0)
        model.fs.unit.inlet.temperature.fix(298.0)
        model.fs.unit.inlet.flow_mol.fix(10)

        return model

    @pytest.mark.unit
    def test_build_model(self, model):

        assert hasattr(model.fs.thermo_params, "component_list")
        assert len(model.fs.thermo_params.component_list) == 3
        assert "H2O" in model.fs.thermo_params.component_list
        assert isinstance(model.fs.thermo_params.H2O, Solvent)
        assert "H_+" in model.fs.thermo_params.component_list
        assert isinstance(model.fs.thermo_params.component("H_+"), Cation)
        assert "OH_-" in model.fs.thermo_params.component_list
        assert isinstance(model.fs.thermo_params.component("OH_-"), Anion)

        assert hasattr(model.fs.thermo_params, "phase_list")
        assert len(model.fs.thermo_params.phase_list) == 1
        assert isinstance(model.fs.thermo_params.Liq, AqueousPhase)

    @pytest.mark.unit
    def test_units_inherent(self, model):
        assert_units_consistent(model)

    @pytest.mark.unit
    def test_dof(self, model):
        assert degrees_of_freedom(model) == 0

    @pytest.fixture
    def expected_stats(self, variant):
        expected = {
            Variant.inherent: {
                number_variables: 81,
                number_total_constraints: 28,
                number_unused_variables: 12,
            },
            Variant.equilibrium: {
                number_variables: 75,
                number_total_constraints: 28,
                number_unused_variables: 6
            }
        }
        return expected[variant]

    @pytest.mark.unit
    def test_stats(self, model, expected_stats):
        for check_func, expected_result in expected_stats.items():
            check_result = check_func(model)
            assert check_result == expected_result

    def _do_inherent_reaction_scaling(self, model):
        for i in model.fs.unit.control_volume.inherent_reaction_extent_index:
            scale = value(
                model.fs.unit.control_volume.properties_out[0.0].k_eq[i[1]].expr
            )
            iscale.set_scaling_factor(
                model.fs.unit.control_volume.inherent_reaction_extent[0.0, i[1]],
                10 / scale,
            )
            iscale.constraint_scaling_transform(
                model.fs.unit.control_volume.properties_out[
                    0.0
                ].inherent_equilibrium_constraint[i[1]],
                0.1,
            )
        return model

    def _do_equilibrium_reaction_scaling(self, model):
        for i in model.fs.unit.control_volume.equilibrium_reaction_extent_index:
            scale = value(model.fs.unit.control_volume.reactions[0.0].k_eq[i[1]].expr)
            iscale.set_scaling_factor(
                model.fs.unit.control_volume.equilibrium_reaction_extent[0.0, i[1]],
                10 / scale,
            )
            iscale.constraint_scaling_transform(
                model.fs.unit.control_volume.reactions[0.0].equilibrium_constraint[
                    i[1]
                ],
                0.1,
            )
        return model

    @pytest.fixture
    def model_scaling(self, model, variant: Variant):
        map_variant_reaction_scale_func = {
            Variant.equilibrium: self._do_equilibrium_reaction_scaling,
            Variant.inherent: self._do_inherent_reaction_scaling,
        }

        scale_func = map_variant_reaction_scale_func[variant]
        scale_func(model)

        # Next, try adding scaling for species
        min = 1e-3
        for i in model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp:
            # i[0] = phase, i[1] = species
            if model.fs.unit.inlet.mole_frac_comp[0, i[1]].value > min:
                scale = model.fs.unit.inlet.mole_frac_comp[0, i[1]].value
            else:
                scale = min
            iscale.set_scaling_factor(
                model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp[i[1]],
                10 / scale,
            )
            iscale.set_scaling_factor(
                model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp[
                    i
                ],
                10 / scale,
            )
            iscale.set_scaling_factor(
                model.fs.unit.control_volume.properties_out[0.0].flow_mol_phase_comp[i],
                10 / scale,
            )
            iscale.constraint_scaling_transform(
                model.fs.unit.control_volume.properties_out[
                    0.0
                ].component_flow_balances[i[1]],
                10 / scale,
            )
            iscale.constraint_scaling_transform(
                model.fs.unit.control_volume.material_balances[0.0, i[1]], 10 / scale
            )

        iscale.calculate_scaling_factors(model.fs.unit)
        return model

    @pytest.mark.component
    def test_scaling(self, model_scaling, variant: Variant):
        model = model_scaling

        assert isinstance(model.fs.unit.control_volume.scaling_factor, Suffix)

        assert isinstance(
            model.fs.unit.control_volume.properties_out[0.0].scaling_factor, Suffix
        )

        assert isinstance(
            model.fs.unit.control_volume.properties_in[0.0].scaling_factor, Suffix
        )

        if variant.is_equilibrium:
            # When using equilibrium reactions, there is another set of scaling factors calculated
            assert isinstance(
                model.fs.unit.control_volume.reactions[0.0].scaling_factor, Suffix
            )

    @pytest.fixture
    def state_args(self):
        return {
            "mole_frac_comp": {
                "H2O": 1,
                "H_+": 10 ** -7 / 55.6,
                "OH_-": 10 ** -7 / 55.6,
            },
            "pressure": 101325,
            "temperature": 298,
            "flow_mol": 10,
        }

    @pytest.fixture
    def model_initialize(self, model, state_args, solver):
        def _collect_data_to_check(m):
            return {
                fixed_variables_set: fixed_variables_set(m),
                activated_constraints_set: activated_constraints_set(m)
            }
        data_before = _collect_data_to_check(model)
        model.fs.unit.initialize(state_args=state_args, optarg=solver.options)

        data_after = _collect_data_to_check(model)
        return model, data_before, data_after

    @pytest.mark.component
    def test_initialize(self, model_initialize):
        model, data_before, data_after = model_initialize

        assert degrees_of_freedom(model) == 0

        for key, value_before in data_before.items():
            value_after = data_after[key]
            assert len(value_before) == len(value_after)

    @pytest.fixture
    def model_solve(self, model, solver):
        results = solver.solve(model, tee=True)
        return model, results

    @pytest.mark.component
    def test_solve(self, model_solve):
        _, results = model_solve
        assert results.solver.termination_condition == TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.component
    def test_solution(self, model_solve):
        model, _ = model_solve

        assert pytest.approx(298, rel=1e-5) == value(
            model.fs.unit.outlet.temperature[0]
        )
        assert pytest.approx(10, rel=1e-5) == value(model.fs.unit.outlet.flow_mol[0])
        assert pytest.approx(101325, rel=1e-5) == value(
            model.fs.unit.outlet.pressure[0]
        )

        total_molar_density = (
            value(
                model.fs.unit.control_volume.properties_out[0.0].dens_mol_phase["Liq"]
            )
            / 1000
        )
        assert pytest.approx(55.2336, rel=1e-5) == total_molar_density

        pH = -value(
            log10(model.fs.unit.outlet.mole_frac_comp[0, "H_+"] * total_molar_density)
        )
        pOH = -value(
            log10(model.fs.unit.outlet.mole_frac_comp[0, "OH_-"] * total_molar_density)
        )
        assert pytest.approx(6.9997414, rel=1e-5) == pH
        assert pytest.approx(6.9997414, rel=1e-5) == pOH
        assert pytest.approx(0.99999, rel=1e-5) == value(
            model.fs.unit.outlet.mole_frac_comp[0.0, "H2O"]
        )


class TestPureWaterEDB(TestPureWater):

    @pytest.fixture
    def thermo_config(self, edb):
        base = edb.get_base("default_thermo")
        elements = ["H", "O"]
        components = []
        # Add the components
        for c in edb.get_components(element_names=elements):
            # Should remove these since we don't have a vapor phase
            # NOTE: The model still runs without removing these, but
            #       the 'stats' won't pass their checks due to the addition
            #       of new system variables that they inherently bring
            c.remove("enth_mol_ig_comp")
            c.remove("pressure_sat_comp")
            base.add(c)
            components.append(c.name)
        # Add the reactions
        for r in edb.get_reactions(component_names=components):
            # Set a custom reaction order
            r.set_reaction_order("Liq", {"H2O": 0, "H_+": 1, "OH_-": 1})
            r._data["type"] = "inherent"
            base.add(r)
        return base.idaes_config


    @pytest.fixture
    def water_reaction_config(self, edb):
        elements = ["H", "O"]
        components = [c.name for c in edb.get_components(element_names=elements)]
        base = edb.get_base("reaction")
        # Add the reactions
        for r in edb.get_reactions(component_names=components):
            # Set a custom reaction order
            r.set_reaction_order("Liq", {"H2O": 0, "H_+": 1, "OH_-": 1})
            # Need to remove this to avoid errors when using the generated config
            r.remove_parameter("ds_rxn_ref")
            base.add(r)
        return base.idaes_config



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
import pytest
from pyomo.environ import ConcreteModel
from idaes.core import FlowsheetBlock
from watertap.examples.flowsheets.full_treatment_train.flowsheet_components import desalination
from watertap.examples.flowsheets.full_treatment_train.model_components import property_models
from watertap.examples.flowsheets.full_treatment_train.util import check_build, check_scaling


@pytest.mark.unit
def test_build_and_scale_desalination():
    for has_desal_feed in [True]:
        for is_twostage in [True, False]:
            for has_ERD in [True, False]:
                for RO_type in ['Sep', '0D', '1D']:
                    for RO_base in ['TDS']:
                        for RO_level in ['simple', 'detailed']:
                            kwargs = {'has_desal_feed': has_desal_feed, 'is_twostage': is_twostage, 'has_ERD': has_ERD,
                                      'RO_type': RO_type, 'RO_base': RO_base, 'RO_level': RO_level}

                            if RO_type == 'Sep' and is_twostage:
                                continue  # not a supported combination
                            elif RO_type == 'Sep' and has_ERD:
                                continue  # not a supported combination
                            elif RO_type == '1D' and RO_level == 'simple':
                                continue  # not a supported combination

                            m = ConcreteModel()
                            m.fs = FlowsheetBlock(default={"dynamic": False})
                            property_models.build_prop(m, base=kwargs['RO_base'])

                            check_build(m, build_func=desalination.build_desalination, **kwargs)
                            assert hasattr(m.fs, 'RO')
                            check_scaling(m, scale_func=desalination.scale_desalination, **kwargs)

                            desalination.display_desalination(m, **kwargs)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

# Import Pyomo libraries
from pyomo.common.config import ConfigBlock, ConfigValue, In
from pyomo.environ import (Block,
                           Var,
                           Suffix,
                           NonNegativeReals,
                           Reals,
                           value,
                           units as pyunits)

# Import IDAES cores
import idaes.logger as idaeslog
from idaes.core import (ControlVolume0DBlock,
                        declare_process_block_class,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        MomentumBalanceType,
                        UnitModelBlockData,
                        useDefault)
from idaes.core.util import get_solver
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.initialization import revert_state_vars
from idaes.core.util.tables import create_stream_table_dataframe
import idaes.core.util.scaling as iscale

_log = idaeslog.getLogger(__name__)

@declare_process_block_class("PressureExchanger")
class PressureExchangerData(UnitModelBlockData):
    """
    Standard Pressure Exchanger Unit Model Class:
    - steady state only
    """
    CONFIG = ConfigBlock()

    CONFIG.declare("dynamic", ConfigValue(
        domain=In([False]),
        default=False,
        description="Dynamic model flag - must be False",
        doc="""Indicates whether this model will be dynamic or not,
    **default** = False. Pressure exchangers do not support dynamic behavior."""))
    CONFIG.declare("has_holdup", ConfigValue(
        default=False,
        domain=In([False]),
        description="Holdup construction flag - must be False",
        doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. Pressure exchangers do not have defined volume, thus
    this must be False."""))
    CONFIG.declare("material_balance_type", ConfigValue(
        default=MaterialBalanceType.useDefault,
        domain=In(MaterialBalanceType),
        description="Material balance construction flag",
        doc="""Indicates what type of mass balance should be constructed,
    **default** - MaterialBalanceType.useDefault.
    **Valid values:** {
    **MaterialBalanceType.useDefault - refer to property package for default
    balance type
    **MaterialBalanceType.none** - exclude material balances,
    **MaterialBalanceType.componentPhase** - use phase component balances,
    **MaterialBalanceType.componentTotal** - use total component balances,
    **MaterialBalanceType.elementTotal** - use total element balances,
    **MaterialBalanceType.total** - use total material balance.}"""))
    CONFIG.declare("energy_balance_type", ConfigValue(
        default=EnergyBalanceType.useDefault,
        domain=In(EnergyBalanceType),
        description="Energy balance construction flag",
        doc="""Indicates what type of energy balance should be constructed,
    **default** - EnergyBalanceType.useDefault.
    **Valid values:** {
    **EnergyBalanceType.useDefault - refer to property package for default
    balance type
    **EnergyBalanceType.none** - exclude energy balances,
    **EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
    **EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
    **EnergyBalanceType.energyTotal** - single energy balance for material,
    **EnergyBalanceType.energyPhase** - energy balances for each phase.}"""))
    CONFIG.declare("momentum_balance_type", ConfigValue(
        default=MomentumBalanceType.pressureTotal,
        domain=In(MomentumBalanceType),
        description="Momentum balance construction flag",
        doc="""Indicates what type of momentum balance should be constructed,
    **default** - MomentumBalanceType.pressureTotal.
    **Valid values:** {
    **MomentumBalanceType.none** - exclude momentum balances,
    **MomentumBalanceType.pressureTotal** - single pressure balance for material,
    **MomentumBalanceType.pressurePhase** - pressure balances for each phase,
    **MomentumBalanceType.momentumTotal** - single momentum balance for material,
    **MomentumBalanceType.momentumPhase** - momentum balances for each phase.}"""))
    CONFIG.declare("property_package", ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description="Property package to use for control volume",
        doc="""Property parameter object used to define property calculations,
        **default** - useDefault.
        **Valid values:** {
        **useDefault** - use default package from parent model or flowsheet,
        **PhysicalParameterObject** - a PhysicalParameterBlock object.}"""))
    CONFIG.declare("property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property packages",
        doc="""A ConfigBlock with arguments to be passed to a property block(s)
        and used when constructing these,
        **default** - None.
        **Valid values:** {
        see property package for documentation.}"""))

    def build(self):
        super().build()

        # Pressure exchanger supports only liquid phase
        if self.config.property_package.phase_list != ['Liq']:
            raise ConfigurationError(
                "Pressure exchanger model only supports one liquid phase ['Liq'],"
                "the property package has specified the following phases {}"
                    .format(self.config.property_package.phase_list))

        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        units_meta = self.config.property_package.get_metadata().get_derived_units

        self.efficiency_pressure_exchanger = Var(
            self.flowsheet().config.time,
            initialize=0.95,
            bounds=(1e-6, 1),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc='Pressure exchanger efficiency')

        # Build control volume for high pressure side
        self.high_pressure_side = ControlVolume0DBlock(default={
            "dynamic": False,
            "has_holdup": False,
            "property_package": self.config.property_package,
            "property_package_args": self.config.property_package_args})

        self.high_pressure_side.add_state_blocks(
            has_phase_equilibrium=False)

        self.high_pressure_side.add_material_balances(
            balance_type=self.config.material_balance_type)

        self.high_pressure_side.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=True)

        self.high_pressure_side.deltaP.setub(0)

        @self.high_pressure_side.Expression(
            self.flowsheet().config.time,
            doc='Work transferred to high pressure side fluid (should be negative)')
        def work(b, t):
            return b.properties_in[t].flow_vol * b.deltaP[t]

        # Build control volume for low pressure side
        self.low_pressure_side = ControlVolume0DBlock(default={
            "dynamic": False,
            "has_holdup": False,
            "property_package": self.config.property_package,
            "property_package_args": self.config.property_package_args})

        self.low_pressure_side.add_state_blocks(
            has_phase_equilibrium=False)

        self.low_pressure_side.add_material_balances(
            balance_type=self.config.material_balance_type)

        self.low_pressure_side.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=True)

        self.low_pressure_side.deltaP.setlb(0)

        @self.low_pressure_side.Expression(
            self.flowsheet().config.time,
            doc='Work transferred to low pressure side fluid')
        def work(b, t):
            return b.properties_in[t].flow_vol * b.deltaP[t]

        # Add Ports
        self.add_inlet_port(name='high_pressure_inlet', block=self.high_pressure_side)
        self.add_outlet_port(name='high_pressure_outlet', block=self.high_pressure_side)
        self.add_inlet_port(name='low_pressure_inlet', block=self.low_pressure_side)
        self.add_outlet_port(name='low_pressure_outlet', block=self.low_pressure_side)

        # Performance equations
        @self.Constraint(
            self.flowsheet().config.time,
            doc="Pressure transfer")
        def eq_pressure_transfer(b, t):
            return (b.low_pressure_side.deltaP[t] ==
                    b.efficiency_pressure_exchanger[t] * -b.high_pressure_side.deltaP[t])

        @self.Constraint(
            self.flowsheet().config.time,
            doc="Equal volumetric flow rate")
        def eq_equal_flow_vol(b, t):
            return (b.high_pressure_side.properties_in[t].flow_vol ==
                    b.low_pressure_side.properties_in[t].flow_vol)

        @self.Constraint(
            self.flowsheet().config.time,
            doc="Equal low pressure on both sides")
        def eq_equal_low_pressure(b, t):
            return (b.high_pressure_side.properties_out[t].pressure ==
                    b.low_pressure_side.properties_in[t].pressure)

        @self.low_pressure_side.Constraint(
            self.flowsheet().config.time,
            doc="Isothermal constraint")
        def eq_isothermal_temperature(b, t):
            return b.properties_in[t].temperature == b.properties_out[t].temperature

        @self.high_pressure_side.Constraint(
            self.flowsheet().config.time,
            doc="Isothermal constraint")
        def eq_isothermal_temperature(b, t):
            return b.properties_in[t].temperature == b.properties_out[t].temperature

    def initialize(
            self,
            state_args=None,
            routine=None,
            outlvl=idaeslog.NOTSET,
            solver=None,
            optarg=None):
        """
        General wrapper for pressure exchanger initialization routine

        Keyword Arguments:
            routine : str stating which initialization routine to execute
                        * None - currently no specialized routine for Pressure exchanger unit
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine (default=idaeslog.NOTSET)
            optarg : solver options dictionary object, if None provided an empty
                     dictionary will be used (default=None)
            solver : solver object or string indicating which solver to use during
                     initialization, if None provided the default solver will be used
                     (default = None)
        Returns: None
        """
        # Get loggers
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="properties")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="properties")

        # Set solver and options
        opt = get_solver(solver, optarg)

        # initialize inlets
        flags_low_in = self.low_pressure_side.properties_in.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
            hold_state=True)
        flags_high_in = self.high_pressure_side.properties_in.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
            hold_state=True)
        init_log.info_high("Initialize inlets complete")

        # check that inlets are feasible
        if (value(self.low_pressure_side.properties_in[0].pressure)
                > value(self.high_pressure_side.properties_in[0].pressure)):
            raise ConfigurationError(
                "Initializing pressure exchanger failed because "
                "the low pressure side inlet has a higher pressure "
                "than the high pressure side inlet")
        if (abs(value(self.low_pressure_side.properties_in[0].flow_vol)
                - value(self.high_pressure_side.properties_in[0].flow_vol))
                / value(self.high_pressure_side.properties_in[0].flow_vol)
                > 1e-4):  # flow_vol values are not within 0.1%
            raise ConfigurationError(
                "Initializing pressure exchanger failed because "
                "the volumetric flow rates are not equal for both inlets")
        else:  # volumetric flow is equal, deactivate flow constraint for the solve
            self.eq_equal_flow_vol.deactivate()

        # initialize outlets from inlets and update pressure
        def propogate_state(sb1, sb2):
            state_dict_1 = sb1.define_state_vars()
            state_dict_2 = sb2.define_state_vars()
            for k in state_dict_1.keys():
                if state_dict_1[k].is_indexed():
                    for m in state_dict_1[k].keys():
                        state_dict_2[k][m].value = state_dict_1[k][m].value
                else:
                    state_dict_2[k].value = state_dict_1[k].value

        # low pressure side
        propogate_state(self.low_pressure_side.properties_in[0],
                        self.low_pressure_side.properties_out[0])
        self.low_pressure_side.properties_out[0].pressure = (
                self.low_pressure_side.properties_in[0].pressure.value
                + self.efficiency_pressure_exchanger[0].value
                * (self.high_pressure_side.properties_in[0].pressure.value
                   - self.low_pressure_side.properties_in[0].pressure.value))
        # high pressure side
        propogate_state(self.high_pressure_side.properties_in[0],
                        self.high_pressure_side.properties_out[0])
        self.high_pressure_side.properties_out[0].pressure.value = \
            self.low_pressure_side.properties_in[0].pressure.value
        init_log.info_high("Initialize outlets complete")

        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info("Initialization complete: {}".format(idaeslog.condition(res)))

        # release state of fixed variables
        self.low_pressure_side.properties_in.release_state(flags_low_in)
        self.high_pressure_side.properties_in.release_state(flags_high_in)

        # reactivate volumetric flow constraint
        self.eq_equal_flow_vol.activate()

    def get_costing(self, module=None):
        self.costing = Block()
        module.PressureExchanger_costing(self.costing)

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        # scale variables
        if iscale.get_scaling_factor(self.efficiency_pressure_exchanger) is None:
            # efficiency should always be between 0.1-1
            iscale.set_scaling_factor(self.efficiency_pressure_exchanger, 1)

        # scale expressions
        if iscale.get_scaling_factor(self.low_pressure_side.work) is None:
            sf = iscale.get_scaling_factor(self.low_pressure_side.properties_in[0].flow_vol)
            sf = sf * iscale.get_scaling_factor(self.low_pressure_side.deltaP[0])
            iscale.set_scaling_factor(self.low_pressure_side.work, sf)

        if iscale.get_scaling_factor(self.high_pressure_side.work) is None:
            sf = iscale.get_scaling_factor(self.high_pressure_side.properties_in[0].flow_vol)
            sf = sf * iscale.get_scaling_factor(self.high_pressure_side.deltaP[0])
            iscale.set_scaling_factor(self.high_pressure_side.work, sf)

        # transform constraints
        for t, c in self.low_pressure_side.eq_isothermal_temperature.items():
            sf = iscale.get_scaling_factor(self.low_pressure_side.properties_in[t].temperature)
            iscale.constraint_scaling_transform(c, sf)

        for t, c in self.high_pressure_side.eq_isothermal_temperature.items():
            sf = iscale.get_scaling_factor(self.high_pressure_side.properties_in[t].temperature)
            iscale.constraint_scaling_transform(c, sf)

        for t, c in self.eq_pressure_transfer.items():
            sf = iscale.get_scaling_factor(self.low_pressure_side.deltaP[t])
            iscale.constraint_scaling_transform(c, sf)

        for t, c in self.eq_equal_flow_vol.items():
            sf = iscale.get_scaling_factor(self.low_pressure_side.properties_in[t].flow_vol)
            iscale.constraint_scaling_transform(c, sf)

        for t, c in self.eq_equal_low_pressure.items():
            sf = iscale.get_scaling_factor(self.low_pressure_side.properties_in[t].pressure)
            iscale.constraint_scaling_transform(c, sf)

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
                {
                    "HP Side In" : self.high_pressure_inlet,
                    "HP Side Out" : self.high_pressure_outlet,
                    "LP Side In" : self.low_pressure_inlet,
                    "LP Side Out" : self.low_pressure_outlet,
                },
                time_point=time_point)

    def _get_performance_contents(self, time_point=0):
        t = time_point
        return { "vars" : { "Efficiency" : self.efficiency_pressure_exchanger[t],
                            "HP Side Pressure Change" : self.high_pressure_side.deltaP[t],
                            "LP Side Pressure Change" : self.low_pressure_side.deltaP[t],

                          },
                "exprs" : {
                            "HP Side Mechanical Work" : self.high_pressure_side.work[t],
                            "LP Side Mechanical Work" : self.low_pressure_side.work[t],
                          },
                "params" : {
                           },
               }



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

from pyomo.environ import (ConcreteModel, value, Param, Var, Constraint, Expression, Objective, TransformationFactory,
                           Block, NonNegativeReals, RangeSet, check_optimal_termination,
                           units as pyunits)
from pyomo.network import Arc, SequentialDecomposition
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.core.util.initialization import (propagate_state)
from idaes.generic_models.unit_models import Feed, Product, Mixer
from idaes.generic_models.unit_models.mixer import MomentumMixingType
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslogger

from watertap.unit_models.reverse_osmosis_0D import (ReverseOsmosis0D,
                                                       ConcentrationPolarizationType,
                                                       MassTransferCoefficient,
                                                       PressureChangeType)
from watertap.unit_models.pump_isothermal import Pump
from watertap.core.util.initialization import assert_degrees_of_freedom, assert_no_degrees_of_freedom
import watertap.examples.flowsheets.lsrro.financials as financials
import watertap.property_models.NaCl_prop_pack as props



def main(number_of_stages, water_recovery=None):
    m = build(number_of_stages)
    set_operating_conditions(m)
    initialize(m)
    solve(m)
    print('\n***---Simulation results---***')
    display_system(m)
    display_design(m)
    display_state(m)

    optimize_set_up(m, water_recovery)
    solve(m)
    print('\n***---Optimization results---***')
    display_system(m)
    display_design(m)
    display_state(m)

    return m


def build(number_of_stages=2):
    # ---building model---
    m = ConcreteModel()

    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = props.NaClParameterBlock()
    financials.add_costing_param_block(m.fs)

    m.fs.NumberOfStages = Param(initialize=number_of_stages)
    m.fs.StageSet = RangeSet(m.fs.NumberOfStages)
    m.fs.LSRRO_StageSet = RangeSet(2, m.fs.NumberOfStages)
    m.fs.NonFinal_StageSet = RangeSet(m.fs.NumberOfStages-1)

    m.fs.feed = Feed(default={'property_package': m.fs.properties})
    m.fs.product = Product(default={'property_package': m.fs.properties})
    m.fs.disposal = Product(default={'property_package': m.fs.properties})

    # Add the mixers
    m.fs.Mixers = Mixer(m.fs.NonFinal_StageSet, default={
            "property_package": m.fs.properties,
            "momentum_mixing_type": MomentumMixingType.equality,  # booster pump will match pressure
            "inlet_list": ['upstream', 'downstream']})

    total_pump_work = 0
    # Add the pumps
    m.fs.PrimaryPumps = Pump(m.fs.StageSet, default={"property_package": m.fs.properties})
    for pump in m.fs.PrimaryPumps.values():
        pump.get_costing(module=financials, pump_type="High pressure")
        total_pump_work += pump.work_mechanical[0]

    # Add the equalizer pumps
    m.fs.BoosterPumps = Pump(m.fs.LSRRO_StageSet, default={"property_package": m.fs.properties})
    for pump in m.fs.BoosterPumps.values():
        pump.get_costing(module=financials, pump_type="High pressure")
        total_pump_work += pump.work_mechanical[0]

    # Add the stages ROs
    m.fs.ROUnits = ReverseOsmosis0D(m.fs.StageSet, default={
            "property_package": m.fs.properties,
            "has_pressure_change": True,
            "pressure_change_type": PressureChangeType.calculated,
            "mass_transfer_coefficient": MassTransferCoefficient.calculated,
            "concentration_polarization_type": ConcentrationPolarizationType.calculated})
    for ro_unit in m.fs.ROUnits.values():
        ro_unit.get_costing(module=financials)

    # Add EnergyRecoveryDevice
    m.fs.EnergyRecoveryDevice = Pump(default={"property_package": m.fs.properties})
    m.fs.EnergyRecoveryDevice.get_costing(module=financials, pump_type="Pressure exchanger")
    total_pump_work += m.fs.EnergyRecoveryDevice.work_mechanical[0]

    # additional variables or expressions
    # system water recovery
    m.fs.water_recovery = Var(
            initialize=0.5,
            bounds=(0, 1),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc='System Water Recovery')
    m.fs.eq_water_recovery = Constraint(expr=\
              sum(m.fs.feed.flow_mass_phase_comp[0,'Liq',:]) * m.fs.water_recovery == \
              sum(m.fs.product.flow_mass_phase_comp[0,'Liq',:]) )

    # additional variables or expressions
    product_flow_vol_total = m.fs.product.properties[0].flow_vol
    m.fs.annual_water_production = Expression(
        expr=pyunits.convert(product_flow_vol_total, to_units=pyunits.m ** 3 / pyunits.year)
             * m.fs.costing_param.load_factor)
    m.fs.specific_energy_consumption = Expression(
        expr=pyunits.convert(total_pump_work, to_units=pyunits.kW)
             / pyunits.convert(product_flow_vol_total, to_units=pyunits.m**3 / pyunits.hr))

    # costing
    financials.get_system_costing(m.fs)
    # objective
    m.fs.objective = Objective(expr=m.fs.costing.LCOW)

    # connections

    # Connect the feed to the first pump
    m.fs.feed_to_pump = Arc(source=m.fs.feed.outlet, destination=m.fs.PrimaryPumps[1].inlet)

    # Connect the primary RO permeate to the product
    m.fs.primary_RO_to_product = Arc(source=m.fs.ROUnits[1].permeate, destination=m.fs.product.inlet)

    # Connect the Pump n to the Mixer n
    m.fs.pump_to_mixer = Arc(m.fs.NonFinal_StageSet,
            rule=lambda fs,n : {'source':fs.PrimaryPumps[n].outlet,
                                'destination':fs.Mixers[n].upstream})

    # Connect the Mixer n to the Stage n
    m.fs.mixer_to_stage = Arc(m.fs.NonFinal_StageSet,
            rule=lambda fs,n : {'source':fs.Mixers[n].outlet,
                                'destination':fs.ROUnits[n].inlet})

    # Connect the Stage n to the Pump n+1
    m.fs.stage_to_pump = Arc(m.fs.NonFinal_StageSet,
            rule=lambda fs,n : {'source':fs.ROUnits[n].retentate,
                                'destination':fs.PrimaryPumps[n+1].inlet})

    # Connect the Stage n to the Eq Pump n
    m.fs.stage_to_eq_pump = Arc(m.fs.LSRRO_StageSet,
            rule=lambda fs,n : {'source':fs.ROUnits[n].permeate,
                                'destination':fs.BoosterPumps[n].inlet})

    # Connect the Eq Pump n to the Mixer n-1
    m.fs.eq_pump_to_mixer = Arc(m.fs.LSRRO_StageSet,
            rule=lambda fs,n : {'source':fs.BoosterPumps[n].outlet,
                                'destination':fs.Mixers[n-1].downstream})

    # Connect the Pump N to the Stage N
    last_stage = m.fs.StageSet.last()
    m.fs.pump_to_stage = Arc(source=m.fs.PrimaryPumps[last_stage].outlet,
            destination=m.fs.ROUnits[last_stage].inlet)

    # Connect Final Stage to EnergyRecoveryDevice Pump
    m.fs.stage_to_erd = Arc(source=m.fs.ROUnits[last_stage].retentate,
            destination=m.fs.EnergyRecoveryDevice.inlet)

    # Connect the EnergyRecoveryDevice to the disposal
    m.fs.erd_to_disposal = Arc(source=m.fs.EnergyRecoveryDevice.outlet,
            destination=m.fs.disposal.inlet)

    # additional bounding
    for b in m.component_data_objects(Block, descend_into=True):
        # NaCl solubility limit
        if hasattr(b, 'mass_frac_phase_comp'):
            b.mass_frac_phase_comp['Liq', 'NaCl'].setub(0.26)

    TransformationFactory("network.expand_arcs").apply_to(m)

    return m


def set_operating_conditions(m):

    # ---specifications---
    # parameters
    pump_efi = 0.75  # pump efficiency [-]
    erd_efi = 0.8  # energy recovery device efficiency [-]
    mem_A = 4.2e-12  # membrane water permeability coefficient [m/s-Pa]
    mem_B = 3.5e-8  # membrane salt permeability coefficient [m/s]
    height = 1e-3  # channel height in membrane stage [m]
    spacer_porosity = 0.97  # spacer porosity in membrane stage [-]
    width = 5 # membrane width factor [m]
    area = 100 # membrane area [m^2]
    pressure_atm = 101325  # atmospheric pressure [Pa]

    # feed
    feed_flow_mass = 1*pyunits.kg/pyunits.s
    feed_mass_frac_NaCl = 70.0/1000.0
    feed_temperature = 273.15 + 25

    # initialize feed
    m.fs.feed.pressure[0].fix(pressure_atm)
    m.fs.feed.temperature[0].fix(feed_temperature)
    m.fs.feed.flow_mass_phase_comp[0, 'Liq', 'NaCl'].fix(feed_flow_mass * feed_mass_frac_NaCl)
    m.fs.feed.flow_mass_phase_comp[0, 'Liq', 'H2O'].fix(feed_flow_mass * (1-feed_mass_frac_NaCl))

    # initialize pumps
    for pump in m.fs.PrimaryPumps.values():
        pump.control_volume.properties_out[0].pressure = 75e5
        pump.efficiency_pump.fix(pump_efi)
        pump.control_volume.properties_out[0].pressure.fix()
        iscale.set_scaling_factor(pump.control_volume.work, 1e-3)

    # initialize eq pumps
    for pump in m.fs.BoosterPumps.values():
        pump.efficiency_pump.fix(pump_efi)
        iscale.set_scaling_factor(pump.control_volume.work, 1e-3)

    # initialize stages
    for idx, stage in m.fs.ROUnits.items():
        if idx > m.fs.StageSet.first():
            B_scale = 100.0
        else:
            B_scale = 1.0
        stage.A_comp.fix(mem_A)
        stage.B_comp.fix(mem_B*B_scale)
        stage.channel_height.fix(height)
        stage.spacer_porosity.fix(spacer_porosity)
        stage.area.fix(area/float(idx))
        stage.width.fix(width)
        stage.permeate.pressure[0].fix(pressure_atm)

    # energy recovery device
    m.fs.EnergyRecoveryDevice.efficiency_pump.fix(erd_efi)
    m.fs.EnergyRecoveryDevice.control_volume.properties_out[0].pressure.fix(pressure_atm)
    iscale.set_scaling_factor(m.fs.EnergyRecoveryDevice.control_volume.work, 1e-3)

    # ---scaling---
    m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1, index=('Liq', 'H2O'))
    m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e2, index=('Liq', 'NaCl'))
    iscale.calculate_scaling_factors(m)

    # ---checking model---
    assert_units_consistent(m)
    assert_no_degrees_of_freedom(m)

    print('Feed Concentration = %.1f ppt' % (value(m.fs.feed.flow_mass_phase_comp[0, 'Liq', 'NaCl'])*1000))


def _lsrro_mixer_guess_initializer( mixer, solvent_multiplier, solute_multiplier, optarg ):

    for vname in mixer.upstream.vars:
        if vname == 'flow_mass_phase_comp':
            for time, phase, comp in mixer.upstream.vars[vname]:
                if comp in mixer.config.property_package.solute_set:
                    mixer.downstream.vars[vname][time,phase,comp].value = \
                            solute_multiplier*mixer.upstream.vars[vname][time,phase,comp].value
                elif comp in mixer.config.property_package.solvent_set:
                    mixer.downstream.vars[vname][time,phase,comp].value = \
                            solvent_multiplier*mixer.upstream.vars[vname][time,phase,comp].value
                else:
                    raise RuntimeError(f"Unknown component {comp}")
        else: # copy the state
            for idx in mixer.upstream.vars[vname]:
                mixer.downstream.vars[vname][idx].value = mixer.upstream.vars[vname][idx].value

    mixer.initialize(optarg=optarg)


def do_initialization_pass(m, optarg, guess_mixers):

    # start with the feed
    m.fs.feed.initialize(optarg=optarg)

    propagate_state(m.fs.feed_to_pump)

    last_stage = m.fs.StageSet.last()
    first_stage = m.fs.StageSet.first()
    for stage in m.fs.StageSet:
        m.fs.PrimaryPumps[stage].initialize(optarg=optarg)

        if stage == last_stage:
            propagate_state(m.fs.pump_to_stage)
        else:
            propagate_state(m.fs.pump_to_mixer[stage])
            if guess_mixers:
                _lsrro_mixer_guess_initializer( m.fs.Mixers[stage],
                        solvent_multiplier=0.5, solute_multiplier=0.2, optarg=optarg )
            else:
                m.fs.Mixers[stage].initialize(optarg=optarg)
            propagate_state(m.fs.mixer_to_stage[stage])

        m.fs.ROUnits[stage].initialize(optarg=optarg)

        if stage == first_stage:
            propagate_state(m.fs.primary_RO_to_product)
        else:
            propagate_state(m.fs.stage_to_eq_pump[stage])
            m.fs.BoosterPumps[stage].initialize(optarg=optarg)
            propagate_state(m.fs.eq_pump_to_mixer[stage])

        if stage == last_stage:
            propagate_state(m.fs.stage_to_erd)
        else:
            propagate_state(m.fs.stage_to_pump[stage])

    # for the end stage
    propagate_state(m.fs.erd_to_disposal)


def do_backwards_initialization_pass(m, optarg):

    first_stage = m.fs.StageSet.first()
    for stage in reversed(m.fs.NonFinal_StageSet):
        m.fs.Mixers[stage].initialize(optarg=optarg)
        propagate_state(m.fs.mixer_to_stage[stage])
        m.fs.ROUnits[stage].initialize(optarg=optarg)
        propagate_state(m.fs.stage_to_pump[stage])
        if stage == first_stage:
            propagate_state(m.fs.primary_RO_to_product)
        else:
            propagate_state(m.fs.stage_to_eq_pump[stage])
            m.fs.BoosterPumps[stage].initialize(optarg=optarg)
            propagate_state(m.fs.eq_pump_to_mixer[stage])


def initialize(m, verbose=False, solver=None):

    # ---initializing---
    # set up solvers
    if solver is None:
        solver = get_solver()

    optarg = solver.options
    do_initialization_pass(m, optarg=optarg, guess_mixers=True)
    for _ in range(m.fs.NumberOfStages.value//2):
        do_backwards_initialization_pass(m, optarg=optarg)
        do_initialization_pass(m, optarg=optarg, guess_mixers=False)

    # set up SD tool
    seq = SequentialDecomposition()
    seq.options.tear_method = "Wegstein"
    seq.options.iterLim = m.fs.NumberOfStages
    seq.options.tear_set = list(m.fs.eq_pump_to_mixer.values())
    seq.options.log_info = True

    # run SD tool
    def func_initialize(unit):
        outlvl = idaeslogger.INFO if verbose else idaeslogger.CRITICAL
        unit.initialize(optarg=solver.options, outlvl=outlvl)
    seq.run(m, func_initialize)


def solve(m, solver=None, tee=False, raise_on_failure=False):
    # ---solving---
    if solver is None:
        solver = get_solver()

    results = solver.solve(m, tee=tee)
    if check_optimal_termination(results):
        return m
    msg = "The current configuration is infeasible. Please adjust the decision variables."
    if raise_on_failure:
        raise RuntimeError(msg)
    else:
        print(msg)
        return None


def optimize_set_up(m, water_recovery=None):

    for pump in m.fs.PrimaryPumps.values():
        pump.control_volume.properties_out[0].pressure.unfix()
        pump.control_volume.properties_out[0].pressure.setlb(10e5)
        pump.control_volume.properties_out[0].pressure.setub(85e5)
        pump.deltaP.setlb(0)

    # unfix eq pumps
    for pump in m.fs.BoosterPumps.values():
        pump.control_volume.properties_out[0].pressure.unfix()
        pump.control_volume.properties_out[0].pressure.setlb(10e5)
        pump.control_volume.properties_out[0].pressure.setub(85e5)
        pump.deltaP.setlb(0)

    # unfix stages
    for idx, stage in m.fs.ROUnits.items():
        stage.area.unfix()
        stage.width.unfix()
        stage.area.setlb(1)
        stage.area.setub(1000)
        stage.width.setlb(0.1)
        stage.width.setub(100)
        stage.N_Re_io[0, 'in'].unfix()

        if idx > m.fs.StageSet.first():
            stage.B_comp.unfix()
            stage.B_comp.setlb(3.5e-8)
            stage.B_comp.setub(3.5e-8 * 1e3)

    min_avg_flux = 1  # minimum average water flux [kg/m2-h]
    min_avg_flux = min_avg_flux / 3600 * pyunits.kg / pyunits.m**2 / pyunits.s  # [kg/m2-s]

    # additional constraints
    if water_recovery is not None:
        m.fs.water_recovery.fix(water_recovery) # product mass flow rate fraction of feed [-]

    # ---checking model---
    assert_units_consistent(m)
    assert_degrees_of_freedom(m, 4 * m.fs.NumberOfStages - (1 if (water_recovery is None) else 2))

    return m


def display_design(m):
    print('--decision variables--')
    for stage in m.fs.StageSet:
        print('Stage %d operating pressure %.1f bar' % (stage, m.fs.ROUnits[stage].inlet.pressure[0].value/1e5))
        print('Stage %d membrane area      %.1f m2'  % (stage, m.fs.ROUnits[stage].area.value))
        print('Stage %d salt perm. coeff.  %.1f LMH' % (stage, m.fs.ROUnits[stage].B_comp[0,'NaCl'].value*(1000.*3600.)))


def display_state(m):
    print('--------state---------')

    def print_state(s, b):
        flow_mass = sum(b.flow_mass_phase_comp[0, 'Liq', j].value for j in ['H2O', 'NaCl'])
        mass_frac_ppm = b.flow_mass_phase_comp[0, 'Liq', 'NaCl'].value / flow_mass * 1e6
        pressure_bar = b.pressure[0].value / 1e5
        print(s.ljust(20) + ': %.3f kg/s, %.0f ppm, %.1f bar' % (flow_mass, mass_frac_ppm, pressure_bar))

    print_state('Feed', m.fs.feed.outlet)
    last_stage = m.fs.StageSet.last()
    first_stage = m.fs.StageSet.first()
    for stage in m.fs.StageSet:

        print_state(f'Primary Pump {stage} out', m.fs.PrimaryPumps[stage].outlet)
        if stage == last_stage:
            pass
        else:
            print_state(f'Mixer {stage} recycle', m.fs.Mixers[stage].downstream)
            print_state(f'Mixer {stage} out', m.fs.Mixers[stage].outlet)

        print_state(f'RO {stage} permeate', m.fs.ROUnits[stage].permeate)
        print_state(f'RO {stage} retentate', m.fs.ROUnits[stage].retentate)
        wr = m.fs.ROUnits[stage].recovery_vol_phase[0, 'Liq'].value
        sr = m.fs.ROUnits[stage].rejection_phase_comp[0, 'Liq', 'NaCl'].value
        print(f"Stage {stage} Volumetric water recovery: {wr*100:.2f}%, Salt rejection: {sr*100:.2f}%")

        if stage == first_stage:
            pass
        else:
            print_state(f'Booster Pump {stage} out', m.fs.BoosterPumps[stage].outlet)

    print_state(f'Disposal', m.fs.disposal.inlet)
    print_state(f'Product', m.fs.product.inlet)


def display_system(m):
    print('----system metrics----')
    feed_flow_mass = sum(m.fs.feed.flow_mass_phase_comp[0, 'Liq', j].value for j in ['H2O', 'NaCl'])
    feed_mass_frac_NaCl = m.fs.feed.flow_mass_phase_comp[0, 'Liq', 'NaCl'].value / feed_flow_mass
    print('Feed: %.2f kg/s, %.0f ppm' % (feed_flow_mass, feed_mass_frac_NaCl * 1e6))

    prod_flow_mass = sum(m.fs.product.flow_mass_phase_comp[0, 'Liq', j].value for j in ['H2O', 'NaCl'])
    prod_mass_frac_NaCl = m.fs.product.flow_mass_phase_comp[0, 'Liq', 'NaCl'].value / prod_flow_mass
    print('Product: %.3f kg/s, %.0f ppm' % (prod_flow_mass, prod_mass_frac_NaCl * 1e6))

    print('Volumetric water recovery: %.1f%%' % (value(m.fs.water_recovery) * 100))
    print('Energy Consumption: %.1f kWh/m3' % value(m.fs.specific_energy_consumption))
    print('Levelized cost of water: %.2f $/m3' % value(m.fs.costing.LCOW))


if __name__ == "__main__":
    import sys
    if len(sys.argv) not in (2,3):
        print("Usage 1: python lsrro.py number_of_stages")
        print("Usage 2: python lsrro.py number_of_stages target_water_recovery_fraction")
    elif len(sys.argv) == 2:
        main(int(sys.argv[1]))
    else:
        main(int(sys.argv[1]), float(sys.argv[2]))



import pytest
from _pytest.config import Config

from watertap.edb import ElectrolyteDB
from watertap.edb.commands import _load_bootstrap


class MockDB(ElectrolyteDB):

    def __init__(self, db="foo", **kwargs):
        from mongomock import MongoClient
        self._client = MongoClient()
        self._db = getattr(self._client, db)
        # note: don't call superclass!
        self._database_name = db
        self._server_url = "mock"


def _reset(edb: ElectrolyteDB):
    edb._client.drop_database(edb.database)


@pytest.fixture(scope="module")
def edb(pytestconfig: Config) -> ElectrolyteDB:
    """
    Create and populate an EDB instance
    """
    mock_allowed = not pytestconfig.option.edb_no_mock
    if ElectrolyteDB.can_connect():
        _edb = ElectrolyteDB()
    else:
        if mock_allowed:
            _edb = MockDB()
        else:
            pytest.fail("EDB could not connect to a database instance, but mocking is not allowed")
    _load_bootstrap(_edb)
    yield _edb
    _reset(_edb)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

"""Translator blocks for supported property packages"""

from pyomo.environ import Constraint
from idaes.generic_models.unit_models.translator import Translator
from idaes.core.util.scaling import calculate_scaling_factors, constraint_scaling_transform, get_scaling_factor
from watertap.examples.flowsheets.full_treatment_train.model_components import property_models


def build_tb(m, base_inlet='ion', base_outlet='TDS', name_str=None):
    """
    Build a translator block to convert for the specified base from inlet to outlet.
    """

    if name_str is None:
        name_str = 'tb_' + base_inlet + '_to_' + base_outlet

    if base_inlet not in ['ion', 'salt']:
        raise ValueError('Unexpected property base inlet {base_inlet} for build_tb'
                         ''.format(base_inlet=base_inlet))
    prop_inlet = property_models.get_prop(m, base=base_inlet)

    if base_outlet not in ['TDS']:
        raise ValueError('Unexpected property base outlet {base_outlet} for build_tb'
                         ''.format(base_outlet=base_outlet))
    prop_outlet = property_models.get_prop(m, base=base_outlet)

    # build translator block
    setattr(m.fs, name_str, Translator(default={"inlet_property_package": prop_inlet,
                                                "outlet_property_package": prop_outlet}))
    blk = getattr(m.fs, name_str)

    # scale translator block to get scaling factors
    calculate_scaling_factors(blk)

    # add translator block constraints
    blk.eq_equal_temperature = Constraint(
        expr=blk.inlet.temperature[0]
             == blk.outlet.temperature[0])
    constraint_scaling_transform(
        blk.eq_equal_temperature,
        get_scaling_factor(blk.properties_in[0].temperature))
    blk.eq_equal_pressure = Constraint(
        expr=blk.inlet.pressure[0]
             == blk.outlet.pressure[0])
    constraint_scaling_transform(
        blk.eq_equal_pressure,
        get_scaling_factor(blk.properties_in[0].pressure))

    if base_inlet == 'ion' and base_outlet == 'TDS':
        blk.eq_H2O_balance = Constraint(
            expr=blk.inlet.flow_mass_phase_comp[0, 'Liq', 'H2O']
                 == blk.outlet.flow_mass_phase_comp[0, 'Liq', 'H2O'])
        constraint_scaling_transform(
            blk.eq_H2O_balance,
            get_scaling_factor(blk.properties_out[0].flow_mass_phase_comp['Liq', 'H2O']))

        blk.eq_TDS_balance = Constraint(
            expr=sum(
                blk.inlet.flow_mass_phase_comp[0, 'Liq', j] for j in ['Na', 'Ca', 'Mg', 'SO4', 'Cl'])
                 == blk.outlet.flow_mass_phase_comp[0, 'Liq', 'TDS'])
        constraint_scaling_transform(
            blk.eq_TDS_balance,
            get_scaling_factor(blk.properties_out[0].flow_mass_phase_comp['Liq', 'TDS']))

    elif base_inlet == 'salt' and base_outlet == 'TDS':
        blk.eq_H2O_balance = Constraint(
            expr=blk.inlet.flow_mass_phase_comp[0, 'Liq', 'H2O']
                 == blk.outlet.flow_mass_phase_comp[0, 'Liq', 'H2O'])
        constraint_scaling_transform(
            blk.eq_H2O_balance,
            get_scaling_factor(blk.properties_out[0].flow_mass_phase_comp['Liq', 'H2O']))

        blk.eq_TDS_balance = Constraint(
            expr=sum(
                blk.inlet.flow_mass_phase_comp[0, 'Liq', j] for j in ['NaCl', 'CaSO4', 'MgSO4', 'MgCl2'])
                 == blk.outlet.flow_mass_phase_comp[0, 'Liq', 'TDS'])
        constraint_scaling_transform(
            blk.eq_TDS_balance,
            get_scaling_factor(blk.properties_out[0].flow_mass_phase_comp['Liq', 'TDS']))

    else:
        raise ValueError('Unexpected property base combination for build_tb')

    blk.properties_in[0].mass_frac_phase_comp  # touch for initialization
    blk.properties_out[0].mass_frac_phase_comp


###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

"""
    Simple example of a flowsheet containing an RO separator unit model and
    a simple NaOCl chlorination post-treatment unit model::

        inlet ---> [RO Separator] ---> permeate ---> (Translator) ---> [Chlorination] ---> outlet
                        |
                        |
                        v
                    retentate (i.e., waste)


    NOTE: The 2 unit models use a different set of state_vars. Thus, this will need to be
    resolved with some clever constraint formulation.

    Both inlet and outlet streams use K for temperature and Pa for pressure (no change needed)

    The flow from RO Separator uses kg/s for individual "species" (H2O and TDS)

    The inlet for Chlorination uses a total molar flow rate in mol/s and mole fractions
    of individual species. To make the appropriate conversions, we will have to start
    by making some assumptions about the molecular weight of TDS.

    MW H2O = 18e-3 kg/mol   MW TDS = 58.4e-3 kg/mol
                            (just assume all as NaCl? : MW Na = 23 g/mol MW Cl = 35.4 g/mol)

    Total Molar Flow = [ m.fs.RO.permeate.flow_mass_phase_comp[0, 'Liq', 'H2O']/(MW H2O) +
                            m.fs.RO.permeate.flow_mass_phase_comp[0, 'Liq', 'TDS']/(MW TDS) ]

    Molefraction of Na --> Based on TDS
                =  [m.fs.RO.permeate.flow_mass_phase_comp[0, 'Liq', 'TDS']/(MW TDS)] / (Total Molar Flow)

    Molefraction of Cl = Molefraction of Na (1:1 ratio in the salt)

    Molefraction of H2O --> Whatever is remaining

    ---------- NOTE: This is only an example ---------
"""

from watertap.examples.flowsheets.full_treatment_train.model_components import (
    unit_separator,
    property_models,
)
from watertap.examples.flowsheets.full_treatment_train.flowsheet_components.chemistry.PostTreatment_SimpleNaOCl_Chlorination import (
    build_simple_naocl_chlorination_unit,
    initialize_chlorination_example,
    display_results_of_chlorination,
    simple_naocl_reaction_config,
)

# Import specific pyomo objects
from pyomo.environ import ConcreteModel, Constraint, TransformationFactory

# Import the core idaes objects for Flowsheets and types of balances
from idaes.core import FlowsheetBlock

from idaes.generic_models.unit_models.translator import Translator
from pyomo.network import Arc

from watertap.examples.flowsheets.full_treatment_train.util import (
    solve_with_user_scaling,
    check_dof,
)

from watertap.examples.flowsheets.full_treatment_train.electrolyte_scaling_utils import (
    approximate_chemical_state_args,
    calculate_chemical_scaling_factors,
)

from watertap.examples.flowsheets.full_treatment_train.chemical_flowsheet_util import (
    seq_decomp_initializer,
)

from idaes.core.util import scaling as iscale

from idaes.core.util import get_solver

__author__ = "Austin Ladshaw"

# Get default solver for testing
solver = get_solver()


def build_SepRO_Chlorination_flowsheet(model, mg_per_L_NaOCl_added=0):
    property_models.build_prop(model, base="TDS")
    unit_separator.build_SepRO(model)
    iscale.calculate_scaling_factors(model.fs.RO)
    property_models.specify_feed(model.fs.RO.mixed_state[0], base="TDS")

    total_molar_density = 1 / 18 * 1000  # mol/L
    free_chlorine_added = mg_per_L_NaOCl_added / 74.44 / 1000 * 70900  # mg/L as NaOCl
    total_chlorine_inlet = free_chlorine_added / 70900  # mol/L
    total_molar_density += total_chlorine_inlet

    # May need to change this build interface
    build_simple_naocl_chlorination_unit(
        model, mg_per_L_NaOCl_added=mg_per_L_NaOCl_added
    )

    # Translator inlet from RO and outlet goes to chlorination
    # NOTE: May need to come up with a way to set state_args for Translator for
    #       better convergence behavior. This block seems to be the trouble maker
    #       for the full solve.
    model.fs.RO_to_Chlor = Translator(
        default={
            "inlet_property_package": model.fs.prop_TDS,
            "outlet_property_package": model.fs.simple_naocl_thermo_params,
        }
    )

    # Add constraints to define how the translator will function
    model.fs.RO_to_Chlor.eq_equal_temperature = Constraint(
        expr=model.fs.RO_to_Chlor.inlet.temperature[0]
        == model.fs.RO_to_Chlor.outlet.temperature[0]
    )
    model.fs.RO_to_Chlor.eq_equal_pressure = Constraint(
        expr=model.fs.RO_to_Chlor.inlet.pressure[0]
        == model.fs.RO_to_Chlor.outlet.pressure[0]
    )

    model.fs.RO_to_Chlor.total_flow_cons = Constraint(
        expr=model.fs.RO_to_Chlor.outlet.flow_mol[0]
        == (model.fs.RO_to_Chlor.inlet.flow_mass_phase_comp[0, "Liq", "H2O"] / 18e-3)
        + (model.fs.RO_to_Chlor.inlet.flow_mass_phase_comp[0, "Liq", "TDS"] / 58.4e-3)
    )

    model.fs.RO_to_Chlor.H_con = Constraint(
        expr=model.fs.RO_to_Chlor.outlet.mole_frac_comp[0, "H_+"] == 0
    )
    model.fs.RO_to_Chlor.OH_con = Constraint(
        expr=model.fs.RO_to_Chlor.outlet.mole_frac_comp[0, "OH_-"] == 0
    )
    model.fs.RO_to_Chlor.HOCl_con = Constraint(
        expr=model.fs.RO_to_Chlor.outlet.mole_frac_comp[0, "HOCl"] == 0
    )

    model.fs.RO_to_Chlor.OCl_con = Constraint(
        expr=model.fs.RO_to_Chlor.outlet.mole_frac_comp[0, "OCl_-"]
        == total_chlorine_inlet / total_molar_density
    )

    model.fs.RO_to_Chlor.Cl_con = Constraint(
        expr=model.fs.RO_to_Chlor.outlet.mole_frac_comp[0, "Cl_-"]
        == (model.fs.RO_to_Chlor.inlet.flow_mass_phase_comp[0, "Liq", "TDS"] / 58.4e-3)
        / model.fs.RO_to_Chlor.outlet.flow_mol[0]
    )

    model.fs.RO_to_Chlor.Na_con = Constraint(
        expr=model.fs.RO_to_Chlor.outlet.mole_frac_comp[0, "Na_+"]
        == (model.fs.RO_to_Chlor.inlet.flow_mass_phase_comp[0, "Liq", "TDS"] / 58.4e-3)
        / model.fs.RO_to_Chlor.outlet.flow_mol[0]
        + model.fs.RO_to_Chlor.outlet.mole_frac_comp[0, "OCl_-"]
    )

    model.fs.RO_to_Chlor.H2O_con = Constraint(
        expr=model.fs.RO_to_Chlor.outlet.mole_frac_comp[0, "H2O"]
        == 1
        - sum(
            model.fs.RO_to_Chlor.outlet.mole_frac_comp[0, j]
            for j in ["H_+", "OH_-", "HOCl", "OCl_-", "Cl_-", "Na_+"]
        )
    )

    # Add the connecting arcs
    model.fs.S1 = Arc(
        source=model.fs.RO.permeate, destination=model.fs.RO_to_Chlor.inlet
    )
    model.fs.S2 = Arc(
        source=model.fs.RO_to_Chlor.outlet, destination=model.fs.simple_naocl_unit.inlet
    )
    TransformationFactory("network.expand_arcs").apply_to(model)

    # Inlet conditions for RO unit already set from the build function

    # Calculate scaling factors and setup each block
    iscale.calculate_scaling_factors(model.fs.RO_to_Chlor)
    state_args, stoich_extents = approximate_chemical_state_args(
        model.fs.simple_naocl_unit,
        model.fs.simple_naocl_rxn_params,
        simple_naocl_reaction_config,
    )
    calculate_chemical_scaling_factors(
        model.fs.simple_naocl_unit,
        model.fs.simple_naocl_thermo_params,
        model.fs.simple_naocl_rxn_params,
        state_args,
    )

    # initialize each chemical block (REQUIRED)
    initialize_chlorination_example(model.fs.simple_naocl_unit, state_args)

    # unfix inlet conditions for chlorination
    model.fs.simple_naocl_unit.inlet.pressure.unfix()
    model.fs.simple_naocl_unit.inlet.temperature.unfix()
    model.fs.simple_naocl_unit.inlet.flow_mol.unfix()

    model.fs.simple_naocl_unit.inlet.mole_frac_comp[0, "H_+"].unfix()
    model.fs.simple_naocl_unit.inlet.mole_frac_comp[0, "OH_-"].unfix()
    model.fs.simple_naocl_unit.inlet.mole_frac_comp[0, "HOCl"].unfix()
    model.fs.simple_naocl_unit.inlet.mole_frac_comp[0, "Cl_-"].unfix()

    #   Here is where I would generally add NaOCl (not sure how best to handle it here)
    #       May have to remove translator constraints? or put added chlorine in those
    #       constraints? Or add a mixer block before the chlorination?
    model.fs.simple_naocl_unit.inlet.mole_frac_comp[0, "OCl_-"].unfix()
    model.fs.simple_naocl_unit.dosing_rate.unfix()
    model.fs.simple_naocl_unit.inlet.mole_frac_comp[0, "Na_+"].unfix()
    model.fs.simple_naocl_unit.inlet.mole_frac_comp[0, "H2O"].unfix()

    check_dof(model)


def run_SepRO_Chlorination_flowsheet_example():
    model = ConcreteModel()
    model.fs = FlowsheetBlock(default={"dynamic": False})

    # build the flow sheet
    build_SepRO_Chlorination_flowsheet(model, mg_per_L_NaOCl_added=2)

    # Call the sequential decomposition initializer tool
    # NOTE: Generally shouldn't use this, just manually initialize
    seq_decomp_initializer(model)

    # run the full solve
    solve_with_user_scaling(model, tee=True, bound_push=1e-5, mu_init=1e-3)

    model.fs.RO.inlet.display()
    model.fs.RO.permeate.display()
    model.fs.RO.retentate.display()

    model.fs.RO_to_Chlor.inlet.display()
    model.fs.RO_to_Chlor.outlet.display()

    display_results_of_chlorination(model.fs.simple_naocl_unit)

    return model


def run_SepRO_Chlorination_flowsheet_with_outlet_constraint_example():
    model = ConcreteModel()
    model.fs = FlowsheetBlock(default={"dynamic": False})

    # build the flow sheet (if constraining outlet, becareful of your initial guess for
    #       the added NaOCl. Scaling is dependent on good guesses)
    build_SepRO_Chlorination_flowsheet(model, mg_per_L_NaOCl_added=3)

    # Call the sequential decomposition initializer tool
    seq_decomp_initializer(model)

    # Redefine the constraints and fixed vars here
    # Here we fix the exit free chlorine then remove the
    #   constraint the defines the amount of OCl from the translator
    #   block to the chlorination process. We are essentially using
    #   the translator block as both a translator and a mixer.
    model.fs.simple_naocl_unit.free_chlorine.fix(2)
    model.fs.RO_to_Chlor.OCl_con.deactivate()

    # run the full solve
    solve_with_user_scaling(model, tee=True, bound_push=1e-5, mu_init=1e-3)

    model.fs.RO.inlet.display()
    model.fs.RO.permeate.display()
    model.fs.RO.retentate.display()

    model.fs.RO_to_Chlor.inlet.display()
    model.fs.RO_to_Chlor.outlet.display()

    display_results_of_chlorination(model.fs.simple_naocl_unit)

    return model


if __name__ == "__main__":
    model = run_SepRO_Chlorination_flowsheet_example()
    model = run_SepRO_Chlorination_flowsheet_with_outlet_constraint_example()
